{-----------------------------------------------------------------------------}
{ Initializing assembly language compiler.                       (c) 1992 YGP }

{$define BIN_only} { only BIN files are implemented }

  (*) segment (*)
  procedure initialize;
                                                                            (*)
    {$I gbl_init.pas }                     { globals initialization procedure }
    {$I lex_init.pas }            { lexical analisys initialization procedure }
    {$I asm_init.pas }        { assembly languge compiler initialization part }
    {$I lab_init.pas }                    { label handler initialization part }
                                                                            (*)
    function str2heap((*)const(*) s: string): pointer;
      var p: stringp;
    begin { str2heap }
      (*) getmemwords(p, length(s) div 2 +1); {*) getmem(p, length(s) +1); {*}
      p^:= s; str2heap:= p;
    end { str2heap };

    function path2heap(var s: string): stringp;
    begin { path2heap }
      if (s <> '') and (not (s[length(s)] in [':', '\'])) then begin
        insert('\', s, length(s)+1);
      end { if };
      path2heap:= str2heap(s);
    end { path2heap };

    const left_1 =   'MC68HC11 Assembler v5.12.';
          middle_1 = '(c) 1992, 1998 YGP software';
          right_1 =  'by YGP';
          left_2 =   '';
          middle_2 = '';
          right_2 =  '';

          cfg_f_name = 'MC68HC11.CFG';                     { config file name }

    procedure pr_header((*)const(*) l, m, r: string);
      var half, ls, ms, rs: integer;
    begin { pr_header }
      ls:= length(l); ms:= length(m); rs:= length(r);
      if ls + ms + rs > 0 then begin        {       }
        half:= (MsgLimit -1 -ls -ms -rs) div 2;
        write(message, #13);
        if half < 0 then
          writeln(l, ' ', m, ' ', r)
        else begin
          writeln(l, m: half + ms, r: half + rs);
        end { if };
      end { if };
    end { pr_header };

    procedure HelpUser;
    begin { HelpUser }
	      {....:...10....:...20....:...30....:...40....:...50....:...60...}
      writeln('usage: _68 [options] filename [options]');
      writeln('  /Dxxx   - define conditionals;');
      writeln('  /Ixxx   - paths for include files;');
      (*)
      writeln('  /Txxx   - path for temporary file;');
      (*)
      writeln('  /Bxxx   - path for .bin file;');
      {$ifNdef BIN_only}
      writeln('  /Pxxx   - path for .pgm file;');
      writeln('  /Cxxx   - path for .cmd file;');
      writeln('  /Oxxx   - path for .obj file;');
      {$endif}
      writeln('  /E[xxx] - error messages filename (including path);');
      writeln('  /L[xxx] - listing filename (including path);');
      writeln('  /X[xxx] - print cross reference;');
      (*)
      writeln('  /W[+|-] - wait for key after an error;');
      (*)
      {$ifNdef BIN_only}
      writeln('  .O[BJ]  - generate .OBJ file;');
      writeln('  .P[GM]  - generate .PGM file;');
      writeln('  .C[MD]  - generate .CMD file;');
      writeln('  .B[IN]  - generate .BIN file;');
      {$endif}
      writeln('  /Q-     - quiet compilation off (termometer style)');
      writeln('  /Q+     - quiet compilation on (filenames and numbers)');
      writeln('  /Q*     - quiet compilation (filenames only)');
      writeln('  @xxx    - use config file xxx (including path).');

      c_flags:= c_flags and (_c_terminated xor word($ffff));
      halt(7);
    end { HelpUser };

    var pcnt:      cardinal;                  { parameter/config line counter }
        cfg_name:  string;                                  { config filename }
        param:     string;                                { program parameter }
        file2asm:  string;                   {      }
        no_source: boolean;      {      source file }
        cfg_file:  text;

    procedure ParamErr;
      var i, p: natural;
    begin { ParamErr }
      write(message, #13(*)#24);{*)); clreol; {*}
      if cfg_name <> '' then
        writeln('Error in config file: ', cfg_name, ', line ', pcnt)
      else begin
	p:= 1;			   {    ,     }
	for i:= 1 to ParamCount do begin       {     }
          ParamStr(i, param);                                  {   }
          write(param, ' ');                             { param used as work }
          if i < pcnt then inc(p, length(param)+1);
	end { for i };
	writeln; writeln('^': p);
	writeln('Invalid Command Line Parameter!');
	writeln;
      end { if };

      HelpUser;
    end { ParamErr };

    const switchars = '/-.';

    procedure define_conditional;
      var lid: identifier;
    begin { define_conditional }
      lid.id_type:= id_defined; pack_size:= pack_ident(idnm, pack_name);
      if (insert_lab(pack_size, pack_name, lid, cc_defined) = nil) or
         (insert_lab(pack_size, pack_name, lid, cc_def_cfg) = nil) then
        unimplemented('no space for define');
    end { define_conditional };

    procedure ProcessParameter;
      var i, wrk: natural; fch, ch: char; ErrorWait: boolean;
    begin { ProcessParameter }
      repeat i:= pos(',', param);               {   ','  ';' }
        if i <> 0 then begin
          param[i]:= ';';
        end { if };
      until i = 0;

      i:= 0; wrk:= 3;    { remove all consecutive space and replace them with }
      repeat i:= pos(' ', param);                                { a single ; }
        if i = 0 then begin
          i:= pos(#9, param);
        end { if };
        if i <> 0 then begin
          if i = wrk then
            delete(param, i, 1)
          else begin
            param[i]:= ';'; wrk:= succ(i);
          end { if };
        end { if };
      until i = 0;

      if param = '' then (*) return {*) exit {*};             { Nothing to do }

      fch:= param[1];
      if (fch = '.') and (length(param) > 1) and
         (param[2] in ['.', '\']) then begin
        fch:= #0;
      end { if };

      if pos(fch, switchars) <> 0 then begin        {   switch char }
        fch:= param[1]; delete(param, 1, 1);             {  switch char }
        ch:= UpCase(param[1]);
        delete(param, 1, 1);                                   {  ch- }
        if (param = '') and
	   ((fch <> '.') and not (ch in ['L', 'E', 'W', 'X', 'Q']) (*)or
            (fch =  '.') and not (ch in ['O', 'P', 'B', 'C'])(*)) then ParamErr;

        case ch of
          'Q': begin
            if (param = '') or (param = '+') then begin
              talk_style:= talkfnames;
              c_flags:= c_flags and not (_c_talkmem); end
            else if (param = '-') then begin
              talk_style:= termometer;
              c_flags:= c_flags or _c_talkmem; end
            else if (param = '*') then begin
              talk_style:= quiet;
              c_flags:= c_flags and not (_c_talkmem);
            end { if };
          end { when quiet compilation };

          'D': begin        { /Dxxx define symbol for conditional compilation }
            for i:= 1 to length(param) do begin
              if not (param[i] in ['A'..'Z','a'..'z','0'..'9','_']) then begin
                ParamErr;
              end { if };
            end { for };
            if length(param) > 15 then param[0]:= #15;
            idnm:= param; UpCaseS(idnm); define_conditional;
          end { when D };

          (*)
          'T': t_path:= path2heap(param);     { /Txxx path for temporary file }
          (*)

          {$ifNdef BIN_only}
          'P': if fch <> '.' then
                 p_path:= path2heap(param)         { /Pxxx path for .PGM file }
               else
                 c_flags:= c_flags or _c_pgm_file and
                 word(word(_c_bin_file + _c_cmd_file + _c_obj_file) xor $ffff);
          'O': if fch <> '.' then
                 o_path:= path2heap(param)         { /Oxxx path for .OBJ file }
               else
                 c_flags:= c_flags or _c_obj_file and
                 word(word(_c_bin_file + _c_cmd_file + _c_pgm_file) xor $ffff);
          'C': if fch <> '.' then
                 c_path:= path2heap(param)         { /Cxxx path for .CMD file }
               else
                 c_flags:= c_flags or _c_cmd_file and
                 word(word(_c_bin_file + _c_pgm_file + _c_obj_file) xor $ffff);
          {$endif}
          'B': if fch <> '.' then
                 b_path:= path2heap(param)         { /Bxxx path for .BIN file }
               else
                 c_flags:= c_flags or _c_bin_file and
                 word(word(_c_pgm_file + _c_cmd_file + _c_obj_file) xor $ffff);

          'I': begin                          { /Ixxx paths for include files }
            insert(';', param, length(param)+1); {  ;    }
            i:= 1;
            while i <= length(param) do begin
              if (param[i] = ';') and (i <> 1) and
                 not (param[i-1] in [';', '\', ':']) then begin
                insert('\', param, i);             {   '\' }
              end { if };
              inc(i);
            end { while };
            if i_path = nil then
              i_path:= str2heap(param)
            else begin
              insert(i_path^, param, length(param)+1);
              (*) freememwords(i_path, length(i_path^) div 2 +1);
              {*) freemem(i_path, length(i_path^) +1); {*}
              i_path:= str2heap(param)
            end { if };
          end { when I };

          'E': begin                           { /E[xxx] - set error filename }
            c_flags:= c_flags or _c_gen_err;
            if param <> '' then begin
              if err_fname <> nil then
                ParamErr
              else begin
                err_fname:= str2heap(param)
              end { if };
            end { if };
          end { when E };

          'L': begin                         { /L[xxx] - set listing filename }
            l_flags:= l_flags or _l_gen_lst;
            if param <> '' then begin
              if lst_fname <> nil then
                ParamErr
              else begin
                lst_fname:= str2heap(param)
              end { if };
            end { if };
          end { when L };

	  'X': begin
	    l_flags:= l_flags or _l_gen_xref;
            if param <> '' then begin
	      if xrf_fname <> nil then
                ParamErr
              else begin
		xrf_fname:= str2heap(param)
              end { if };
            end { if };
	  end { when X };

          'W': begin                            { wait for key after an error }
            ErrorWait:= (param = '+') or (param = '');
            if not ErrorWait and (param <> '-') then begin
              ParamErr;
            end { if };
            if ErrorWait then
              c_flags:= c_flags or _c_err_wait
            else begin
              c_flags:= c_flags and (_c_err_wait xor word($ffff));
            end { if };
          end { when W };
        else
          ParamErr;
        end { case } end
      else if no_source then begin    {         }
        file2asm:= param;
        no_source:= false;
        i:= length(param);  {        .  \ }
        wrk:= i +ScanFor('\', param, i, -i);   {       }
        if i-wrk = ScanFor('.', param, wrk+1, i - wrk) then
          insert('.asm', param, i+1)
        else begin
          i:= ScanFor('.', file2asm, wrk+1, i - wrk);
          delete(file2asm, wrk+1+i, length(file2asm) -wrk -i);
        end { if };
        include_source(param, nil); end
      else begin
        writeln('*** Too many files to assembly');
        ParamErr;                      {       }
      end { if };
    end { ProcessParameter };

    var _h, _m, _s, _s100: word;

  begin { initialize }                                                   (*){*)
    save_exit:= exitproc; exitproc:= @comp_exit;     (* Turbo Pascal (EXIT:) *)
    hide_cursor;                                                            {*}
    {$ifdef with_cursor} (*){*)show_cursor;{*} {$endif}

    {$ifdef BIN_only}
    c_flags:= c_flags or _c_bin_file;
    {$endif}

    GetTime(_h, _m, _s, _s100);                     { get compiler start time }
    with start_time do begin     {          }
      h:= _h; m:= _m; s:= _s; s100:= _s100;   {   , ,  }
    end { with };

    pr_header(left_1, middle_1, right_1);
    pr_header(left_2, middle_2, right_2);

    gbl_init;                    { call all initialization parts of assembler }
    lex_init;
    asm_init;
    lab_init;

    no_source:= true; err_fname:= nil; lst_fname:= nil; code_fname:= nil;
    xrf_fname:= nil;
                       { Get parameters from config file and/or command line. }
    cfg_name:= '';                         { Get parameters from config file. }
    for pcnt:= 1 to ParamCount do begin    { Search for @xxx in command line. }
      ParamStr(pcnt, param);
      if param[1] = '@' then begin        { get last parameter with leading @ }
        cfg_name:= param;
      end { if };
    end { for };

    if cfg_name <> '' then begin
      delete(cfg_name, 1, 1);                              { remove leading @ }
      if cfg_name <> '' then begin
        (*) reset(cfg_file, cfg_name);
        {*) filemode:= 0; assign(cfg_file, cfg_name); reset(cfg_file); {*}
        if IOresult <> 0 then begin
          writeln('*** Config file: ', cfg_name, ' not found!'); HelpUser;
        end { if };
      end { if }; end
    else begin
      cfg_name:= cfg_f_name;      { search for this file in current directory }
      (*) reset(cfg_file, cfg_name);
      {*) filemode:= 0; assign(cfg_file, cfg_name); reset(cfg_file); {*}
      if IOresult <> 0 then begin             {        }
        ParamStr(0, param);                { exec dir-. Get path of compiler }
        pcnt:= length(param)+ScanFor('\', param, length(param),-length(param));
        if pcnt <> 0 then                   {     }
          param[0]:= chr(pcnt)                               {   }
        else begin
          param[0]:= chr(pos(':', param));           {   }
        end { if };
        if param = '' then   {   ''         }
          cfg_name:= ''
	else begin
          insert(param, cfg_name, 1);
          (*) reset(cfg_file, cfg_name);
          {*) filemode:= 0; assign(cfg_file, cfg_name); reset(cfg_file); {*}
          if IOresult <> 0 then begin
            cfg_name:= '';                            { no config file at all }
          end { if };
	end { if }
      end { if };
    end { if };

    if cfg_name <> '' then begin
      pcnt:= 0;
      while not eof(cfg_file) do begin
        readln(cfg_file, param); inc(pcnt);
                                                        { skip leading spaces }
        while (param <> '') and (param[1] in [' ', #9]) do begin
          delete(param, 1, 1);
        end { while };
                                { comment lines (begining with ;/*/#) skipped }
        if (param <> '') and not(param[1] in [';', '*', '#']) then begin
          if (param <> '') and (pos(param[1], switchars) <> 0) then
            ProcessParameter
          else begin
            ParamErr;
          end { if };
        end { if };
      end { while };
      close(cfg_file); (*) {*) if IOresult <> 0 then; {*}
    end { if };
    cfg_name:= '';                                    { config file processed }

    for pcnt:= 1 to ParamCount do begin   { Get parameters from command line. }
      ParamStr(pcnt, param);
      if param[1] <> '@' then begin
        ProcessParameter;
      end { if };
    end { for };

    if no_source then HelpUser;                  { print help screen and quit }
    if (c_flags and _c_gen_err <> 0) and (err_fname = nil) then begin
      insert('.err', file2asm, length(file2asm)+1);
      err_fname:= str2heap(file2asm); dec(file2asm[0], 4);
      (*) rewrite(err_file, err_fname^);
      {*) assign(err_file, err_fname^); rewrite(err_file); {*}
    end { if };
    if (l_flags and _l_gen_lst <> 0) and (lst_fname = nil) then begin
      insert('.lst', file2asm, length(file2asm)+1);
      lst_fname:= str2heap(file2asm); dec(file2asm[0], 4);
    end { if };
    if (l_flags and _l_gen_xref <> 0) and (xrf_fname = nil) then begin
      insert('.xrf', file2asm, length(file2asm)+1);
      xrf_fname:= str2heap(file2asm); dec(file2asm[0], 4);
    end { if };
    c_flags:= c_flags or _c_inited;                    { compiler initialized }
  end { initialize };

{-----------------------------------------------------------------------------}
